
# 
# Camelcase from https://gist.github.com/jaytaylor/3660565
#

import re

_underscorer1 = re.compile(r'(.)([A-Z][a-z]+)')
_underscorer2 = re.compile('([a-z0-9])([A-Z])')
 
def camelToSnake(s):
    """
        Is it ironic that this function is written in camel case, yet it
        converts to snake case? hmm..
    """
    subbed = _underscorer1.sub(r'\1_\2', s)
    return _underscorer2.sub(r'\1_\2', subbed).lower()





#
# py.test based tests
#

import dsltool
from inspect import cleandoc

import pytest

#
# DSL Objects
#


class AddDsl(object):
    
    _items = []
    
    @property 
    def items(self):
        return dsltool.add_to_list(self._items)
    
    @items.setter
    def items(self, value):
        self._items = value
        

class ChildDsl(object):
    child_var = None

class ListOfDsl(object):
    var = None
    
    def __init__(self):
        self.child_dsl = dsltool.list_of(ChildDsl)

#
# Test add_to_list
#

def test_addto_0():
    dsl_contents = cleandoc('''
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, AddDsl)
    assert result._items == []

def test_addto_1():
    dsl_contents = cleandoc('''
        items += 1
        items += 2 
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, AddDsl)
    assert result._items == [1,2]
    
#
# Test list_of
#

def test_listof_0():
    dsl_contents = cleandoc('''
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, ListOfDsl)
    assert result.child_dsl == []

def test_listof_1a():
    dsl_contents = cleandoc('''
        with ChildDsl:
            pass
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, ListOfDsl)
    assert len(result.child_dsl) == 1
    assert type(result.child_dsl[0]) is ChildDsl
    
def test_listof_1b():
    dsl_contents = cleandoc('''
        with ChildDsl():
            pass
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, ListOfDsl)
    assert len(result.child_dsl) == 1
    assert type(result.child_dsl[0]) is ChildDsl

def test_listof_1c():
    dsl_contents = cleandoc('''
        with ChildDsl():
            child_var = 1
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, ListOfDsl)
    assert len(result.child_dsl) == 1
    assert type(result.child_dsl[0]) is ChildDsl
    assert result.child_dsl[0].child_var == 1

def test_listof_2a():
    dsl_contents = cleandoc('''
        with ChildDsl:
            child_var = 0
        with ChildDsl:
            child_var = 1
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, ListOfDsl)
    assert len(result.child_dsl) == 2
    for i, child in enumerate(result.child_dsl):
        assert type(child) is ChildDsl
        assert child.child_var == i

def test_listof_2b():
    dsl_contents = cleandoc('''
        with ChildDsl:
            child_var = 0
        with ChildDsl:
            child_var = 1
        with ChildDsl:
            child_var = 2
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, ListOfDsl)
    assert len(result.child_dsl) == 3
    for i, child in enumerate(result.child_dsl):
        assert type(child) is ChildDsl
        assert child.child_var == i

def test_listof_2c():
    dsl_contents = cleandoc('''
        with ChildDsl():
            child_var = 0
        with ChildDsl():
            child_var = 1
        with ChildDsl():
            child_var = 2
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, ListOfDsl)
    assert len(result.child_dsl) == 3
    for i, child in enumerate(result.child_dsl):
        assert type(child) is ChildDsl
        assert child.child_var == i






#
# TODO: Other useful operations
#



class add_to_list(object):
    '''
        Useful for implementing the += operation in your DSL. 
        
        DSL Implementation Object Usage:
        
            class Foo(object):
                
                _items = []
                
                @property
                def items(self):
                    return dsltool.add_to_list(self._items)
                
                @add.setter
                def items(self, value):
                    self._items = value
                
        DSL End User Usage:
        
            with Foo:
                add += 'item'
        
    '''
    
    def __init__(self, lst):
        self.lst = lst
        
    def __iadd__(self, item):
        self.lst.append(item)
        return self.lst
    
    
class list_of(list):
    '''
        When seen in a DSL Implementation Object, this allows the 'with'
        statement to occur multiple times, and each time a new object will
        be appended to the list.
        
        @warning Don't use it like this, or you will get odd errors:
        
            class Foo:
                bar = dsltool.list_of(Bar)
                
        Instead, prefer this:
        
            class Foo:
                def __init__(self):
                    self.bar = dsltool.list_of(Bar)
    '''
    
    def __init__(self, cls):
        list.__init__(self)
        self._original_cls = cls




import inspect

from helpers import list_of
from utils import camelToSnake


class DslExecError(Exception):
    pass

class WithProxy(object):
    '''
        There's probably a better way of doing this.
        
        The __getattribute__ function makes it so we can't actually
        perform accesses on our own object, thus why there are all the
        double underscore function calls to object. :) 
        
        This supports the following:
        
            with Foo:
            with Foo as f:
            with Foo():
            with Foo() as f:
            
        But really, you shouldn't be using Foo() as f, as avoiding that
        is the whole point of this DSL stuff... ;)
    '''
    
    def __init__(self, on_construction, on_exit):
        object.__setattr__(self, 'on_construction', on_construction)
        object.__setattr__(self, 'on_exit', on_exit)
    
    def __call__(self, *args, **kwargs):
        '''This is called at 'with Foo()' '''
        on_construction = object.__getattribute__(self, 'on_construction')
        if on_construction is None:
            return getattr(object.__getattribute__(self, 'constructed_obj'), '__call__')(*args, **kwargs)

        constructed_obj = on_construction(*args, **kwargs)
        object.__setattr__(self, 'constructed_obj', constructed_obj)
        object.__setattr__(self, 'on_construction', None)
            
        return self
    
    #
    # without these 'with Foo() as f:' won't work correctly.
    # 
    
    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, 'constructed_obj'), name)
    
    def __setattr__(self, name, value):
        return setattr(object.__getattribute__(self, 'constructed_obj'), name, value)
    
    def __enter__(self):
        
        # if this is set, the with statement was called
        # without parens. Which is ok, just do the default.

        on_construction = object.__getattribute__(self, 'on_construction')
        if on_construction is not None:            
            constructed_obj = on_construction()
            object.__setattr__(self, 'constructed_obj', constructed_obj)
            object.__setattr__(self, 'on_construction', None)
            
        return self
    
    def __exit__(self, type, value, traceback):
        object.__setattr__(self, 'constructed_obj', None)
        object.__getattribute__(self, 'on_exit')(None)
    

class GlobalsObjectProxy(dict):
    '''
        Used as the 'globals' for an executed DSL file. This is where most
        of the black magic is. 
    '''
        
    def __init__(self, wrapped_obj, filename):
        
        # populate initial dict
        if filename is not None:
            dict.__setitem__(self, '__file__', filename)
        
        self.obj_stack = [wrapped_obj]
        self._pop_stack()
        self._last_push = '<internal error>'
        
    def _pop_stack(self):
        self.wrapped_obj = self.obj_stack.pop()
    
    def _push_stack(self):
        self._last_push = self.wrapped_obj.__class__.__name__
        self.obj_stack.append(self.wrapped_obj)
        self.wrapped_obj = None

    def _check_for_invalid_usage(self):
        if self.wrapped_obj is None:
            raise DslExecError("DSL object name was referenced outside of a 'with' statement (likely was: '%s')" % self._last_push)
            
    def _clean_up_instance(self, inst=None):
        # TODO: refactor
        # TODO: this may be buggy, need to test on properties
        pop_stack = False
        if inst == None:
            pop_stack = True
            inst = self.wrapped_obj
            
        for attr in dir(inst):
            if attr.startswith('_'):
                continue
            
            if inspect.isclass(getattr(inst, attr)):
                setattr(inst, attr, None)
                
        if pop_stack:
            self._pop_stack()
        
            
    def __getitem__(self, k):
        
        #
        # TODO: could have something parse the object before
        # execution, and single out valid/invalid keys, so we
        # would just look in that dict. That way the DSL user couldn't
        # do silly things like overwrite attributes we're looking to
        # use for 'with' statements
        #
        
        self._check_for_invalid_usage()
        
        if hasattr(self.wrapped_obj, k):
            return getattr(self.wrapped_obj, k)
        
        # check for the attribute in snake case. if it exists, 
        # then we are probably using a with statement.
        # 
        # TODO: this can cause subtle bugs, pre-parsing the object
        #       would allow us to catch those
        #
        
        sk = camelToSnake(k)
        
        if hasattr(self.wrapped_obj, sk):
            
            cls = v = getattr(self.wrapped_obj, sk)
            
            # handle properties correctly
            if isinstance(v, property):
                v = getattr(self.wrapped_obj, sk)
            
            attr_is_list = False
            if isinstance(v, list_of):
                attr_is_list = True
                cls = v._original_cls
            
            elif not inspect.isclass(v):
                raise AttributeError("'%s' object has no attribute '%s'" % (self.wrapped_obj.__class__.__name__, k))
            
            # the DSL is asking for a class, return a proxy for the with
            # statement. If they are using list_of, then we append the
            # new instance to the list
            
            # prevent user from calling 'with' twice if they didn't use list_of 
            if not attr_is_list and not inspect.isclass(getattr(self.wrapped_obj, sk)):
                raise AttributeError("'with' statement for '%s' can only be used once per '%s' instance" % (k, self.wrapped_obj.__class__.__name__))
            
            old_object = self.wrapped_obj
            
            # We can't create the child object until the constructor is
            # called, so the WithProxy does that via this function 
            
            def on_construction(*args, **kwargs):
                # if the object is passed an incorrect amount of
                # parameters, you get weird error messages here
                self.wrapped_obj = cls(*args, **kwargs)
                if attr_is_list:
                    getattr(old_object, sk).append(self.wrapped_obj)
                else:
                    setattr(old_object, sk, self.wrapped_obj)
                
                return self.wrapped_obj
            
            self._push_stack()
            return WithProxy(on_construction, self._clean_up_instance)
        
        # return default globals
        return dict.__getitem__(self, k)
            
    
    def __setitem__(self, k, v):
        self._check_for_invalid_usage()
        
        if hasattr(self.wrapped_obj, k):
            setattr(self.wrapped_obj, k, v)
        else:
            dict.__setitem__(self, k, v)
        


def parse_dsl_file(filename, obj):
    '''
        Given a file of code formatted in the DSL that corresponds to the 
        given object, parse the content and store it in the object.
    
        :param filename:  File to read from
        :param cls: Class or object
    '''
    
    with open(filename, 'r') as fp:
        return parse_dsl(fp.read(), obj, filename)


def parse_dsl(string, obj, filename=None):
    '''
        Given a bit of code formatted in the DSL that corresponds to the 
        given object, parse the content and store it in the object.
    
        :param string: String of python code to interpret
        :param cls: Class or object
    '''
    
    if inspect.isclass(obj):
        obj = obj()
    
    # create a wrapper around the object
    dsl_globals = GlobalsObjectProxy(obj, filename)
    
    if filename is None:
        filename = '<string>'
    
    code = compile(string, filename, 'exec')
    
    # TODO: memory leaks warned about in http://lucumr.pocoo.org/2011/2/1/exec-in-python/
    exec code in dsl_globals
    
    # Fix up objects that aren't used
    # -> there's probably a better way to do this
    dsl_globals._clean_up_instance(obj)
    
    # TODO: add 'required' elements to the DSL
       
    return obj
    



from dsltool.utils import camelToSnake


def test_utils():
    '''From https://gist.github.com/jaytaylor/3660565'''
    
    assert camelToSnake('snakesOnAPlane') == 'snakes_on_a_plane'
    assert camelToSnake('SnakesOnAPlane') == 'snakes_on_a_plane'
    assert camelToSnake('snakes_on_a_plane') == 'snakes_on_a_plane'
    assert camelToSnake('IPhoneHysteria') == 'i_phone_hysteria'
    assert camelToSnake('iPhoneHysteria') == 'i_phone_hysteria'
    




#
# py.test based tests
#

import dsltool
from inspect import cleandoc

#
# DSL Objects
#

class BasicDsl(object):
    var = None
    other_var = None


class BasicPropertyDsl(object):
    
    def __init__(self):
        self._var = None
        self.other_var = None
    
    @property
    def var(self):
        return self._var
        
    @var.setter
    def var(self, val):
        self._var = val

#
# Tests
#

def test_basic_0():
    dsl_contents = cleandoc('''
        
    ''')
    
    result1 = dsltool.parse_dsl(dsl_contents, BasicDsl)
    assert result1.var == None
    
    result2 = dsltool.parse_dsl(dsl_contents, BasicPropertyDsl)
    assert result2.var == None

def test_basic_1():
    dsl_contents = cleandoc('''
        var = 3
    ''')
    
    result1 = dsltool.parse_dsl(dsl_contents, BasicDsl)
    assert result1.var == 3
    
    result2 = dsltool.parse_dsl(dsl_contents, BasicPropertyDsl)
    assert result2.var == 3
    
def test_basic_2():
    dsl_contents = cleandoc('''
        var = 3
        other = 4
    ''')
    
    result1 = dsltool.parse_dsl(dsl_contents, BasicDsl)
    assert result1.var == 3
    
    result2 = dsltool.parse_dsl(dsl_contents, BasicPropertyDsl)
    assert result2.var == 3
    
    
def test_basic_3():
    dsl_contents = cleandoc('''
        class Foo(object):
            def get_3(self):
                return 3
        
        foo = Foo()
        var = foo.get_3()
    ''')
    
    result1 = dsltool.parse_dsl(dsl_contents, BasicDsl)
    assert result1.var == 3
    
    result2 = dsltool.parse_dsl(dsl_contents, BasicPropertyDsl)
    assert result2.var == 3
    
def test_basic_4():
    dsl_contents = cleandoc('''
        var = 3
        other_var = var
    ''')
    
    result1 = dsltool.parse_dsl(dsl_contents, BasicDsl)
    assert result1.var == 3
    assert result1.other_var == 3
    
    result2 = dsltool.parse_dsl(dsl_contents, BasicPropertyDsl)
    assert result2.other_var == 3
    



#
# py.test based tests
#

import dsltool
from inspect import cleandoc

import pytest

#
# DSL Objects
#

class GrandchildDsl(object):
    grandchild_var = None

class ChildDsl(object):
    child_var = None
    grandchild_dsl = GrandchildDsl

class BaseDsl(object):
    var = None
    
    child_dsl = ChildDsl


#
# Test nested with statements
#


def test_with_0a0():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl:
            child_var = 4
            with GrandchildDsl:
                grandchild_var = 5
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == 3
    assert result.child_dsl.child_var == 4
    assert result.child_dsl.grandchild_dsl.grandchild_var == 5
    
def test_with_0a1():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl as c:
            c.child_var = 4
            with GrandchildDsl as g:
                g.grandchild_var = 5
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == 3
    assert result.child_dsl.child_var == 4
    assert result.child_dsl.grandchild_dsl.grandchild_var == 5
    
def test_with_0b0():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl():
            child_var = 4
            with GrandchildDsl():
                grandchild_var = 5
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == 3
    assert result.child_dsl.child_var == 4
    assert result.child_dsl.grandchild_dsl.grandchild_var == 5
    
def test_with_0b1():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl() as c:
            c.child_var = 4
            with GrandchildDsl() as g:
                g.grandchild_var = 5
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == 3
    assert result.child_dsl.child_var == 4
    assert result.child_dsl.grandchild_dsl.grandchild_var == 5    

#
# Test missing with statements
#

def test_with_1a():
    dsl_contents = cleandoc('''
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == None
    assert result.child_dsl is None

def test_with_1b():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl:
            pass
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == 3
    assert result.child_dsl.child_var is None
    assert result.child_dsl.grandchild_dsl is None

def test_with_1c0():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl:
            child_var = 4
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == 3
    assert result.child_dsl.child_var == 4
    assert result.child_dsl.grandchild_dsl is None
    
def test_with_1c1():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl as c:
            c.child_var = 4
    ''')
    
    result = dsltool.parse_dsl(dsl_contents, BaseDsl)
    assert result.var == 3
    assert result.child_dsl.child_var == 4
    assert result.child_dsl.grandchild_dsl is None
    
#
# Test multiple with statements
#

def test_with_2a():
    dsl_contents = cleandoc('''
        var = 3
        with ChildDsl:
            child_var = 4
        with ChildDsl:
            child_var = 5
    ''')
    
    with pytest.raises(AttributeError):
        dsltool.parse_dsl(dsl_contents, BaseDsl)
    
#
# Test missing classes
#

def test_with_3a():
    dsl_contents = cleandoc('''
        var = 3
        with MissingClass:
            var = 4
    ''')
    
    with pytest.raises(NameError):
        dsltool.parse_dsl(dsl_contents, BaseDsl)
    



#!/usr/bin/env python

from os.path import join, dirname
from distutils.core import setup

packages = [
    'dsltool',
]

setup(name='pydsltool',
      version='0.1.0',
      description='A library to enable making simple domain specific languages',
      long_description=open(join(dirname(__file__), 'README'), 'r').read(),
      author='Dustin Spicuzza',
      author_email='dustin@virtualroadside.com',
      url='https://github.com/virtuald/pydsltool',
      license='Apache 2.0',
      packages=packages,
      package_dir={'': 'src'},
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development'
      ]
)




#!/usr/bin/env python

import os.path
import sys
import dsltool


#
# When the DSL is executed, the methods and attributes of these objects
# are called
#

class Ingredients(object):
    
    def __init__(self):
        self._items = []
    
    @property
    def add(self):
        '''
            Uses a utility that gives the DSL a syntax to add to a list:
            
                with Ingredients:
                    add += 'item'
                    ...
            
        '''
        return dsltool.add_to_list(self._items)
    
    @add.setter
    def add(self, value):
        self._items = value


class Steps(object):
    '''Example of how to record whatever the user throws at you'''
        
    def __init__(self):
        self._steps = []
    
    def __getattr__(self, name):
        return lambda *args: self._steps.append([name] + list(args))


class Instructions(object):
    preheat_oven = None
    
    #
    # When the DSL is executed, attributes that have a class object assigned 
    # to them are used by the DSL in 'with' statements. The attribute
    # must be a snake cased version of the class name.  
    #
    
    ingredients = Ingredients
    steps = Steps
    
    def __init__(self, name):
        self.name = name
        

class Recipe(object):
    '''This is the base object, so the DSL code is interpreted with this
       object's methods/attributes as its context'''
    
    author = None
    level = None
    
    def __init__(self):
        self.instructions = dsltool.list_of(Instructions)

# 
# Main
#

if __name__ == '__main__':
    
    if len(sys.argv) == 1:
        filename = os.path.join(os.path.dirname(__file__), 'cookies.pydsl')
    else:
        filename = sys.argv[1]
    
    recipe = dsltool.parse_dsl_file(filename, Recipe()) 

    # display the parsed recipe
    print 'Author: %s' % recipe.author
    print 'Level : %s' % recipe.level
    
    for instructions in recipe.instructions:
        print 'Instructions for %s:' % instructions.name
        print '  Preheat oven: %s' % instructions.preheat_oven
        print '  Ingredients:'
        
        if instructions.ingredients is not None:
            for quantity, ingredient in instructions.ingredients._items:
                print '    %s %s' % (quantity, ingredient)
    
        print '  Steps:'
        if instructions.steps is not None:
            for step in instructions.steps._steps:
                print '    %s' % ' '.join(step)
    
    
